#include "Storage.h"
#include "mythread.h"
#include "../BackEnd/backend.h"

#include "../zlib/zlib.h"
#pragma comment(lib,"zlibd.lib")


int CDP_HashInit(CDP_HASH_T *h, int n)
{
	int i;
	
	_TRACE(TRACE_CACHE,"hash_init : n=%d\n",n);
	h->n = n;
	h->insertions = 0;
	h->collisions = 0;
	h->data_len = 0;
	h->parity_len = 0;
	h->archive_len = 0;
	h->total_write_count = 0;
	if ((h->bucket=(PCDP_SUPER_BLOCK)malloc(n*sizeof(CDP_SUPER_BLOCK)))==NULL) {
		TRACE_ERROR("hash_init: malloc1 failed\n");
		return ISCSI_ERROR;
	}
	
	for (i=0; i<n; i++)
	{
		h->bucket[i].next = NULL;
		h->bucket[i].lba = -1;
		h->bucket[i].data_offset = -1;
		h->bucket[i].parity_offset = -1;
		h->bucket[i].last_parity_offset = -1;
#ifndef CDP_NO_GAP
		h->bucket[i].archive_offset = -1;
		h->bucket[i].dirty = 0;
#endif
		h->bucket[i].read_count = 0;
		h->bucket[i].write_count = 0;
	}
	
	return ISCSI_SUCCESS;
}
int CDP_HashClose(CDP_HASH_T *h)
{
	int i;
	PCDP_SUPER_BLOCK curr,prev;
	for(i=0;i<h->n;i++)
	{
		curr = h->bucket[i].next;
		while(curr)
		{
			prev = curr;
			curr = curr->next;
			free(prev);
			TRACE_CDP("free ext %p of %d",prev,i);
		}
	}
	free(h->bucket);
	return ISCSI_SUCCESS;
}
#ifdef CDP_NO_GAP
int CDP_HashInsert(CDP_HASH_T *h, ULONG key, __int64 *data_offset, __int64 *parity_offset, __int64 *last_parity_offset, BOOL rw)
{
	int i;
	PCDP_SUPER_BLOCK ext = NULL;
	PCDP_SUPER_BLOCK ext_prev = NULL;
	int found = 0;
	
	i = key%(h->n);
	_TRACE(TRACE_CACHE,"hash_insert : key=%u, i=%d\n",key,i);
	if(h->bucket[i].lba==(int)key)
	{
		TRACE_CDP("hash_insert : found key=%u, i=%d\n",key,i);
		*data_offset = h->bucket[i].data_offset;
		*parity_offset = h->bucket[i].parity_offset;
		
		if (rw)
		{
			h->bucket[i].write_count++;
			*last_parity_offset = h->bucket[i].last_parity_offset;
			h->bucket[i].last_parity_offset = h->parity_len;
		}
		else
			h->bucket[i].read_count++;
		
		return ISCSI_SUCCESS;
	}
	if (h->bucket[i].lba==-1) {
		TRACE_CDP("inserting key %u  into %d,offset %d\n", key, i,h->data_len);
		h->bucket[i].lba = key;
		h->bucket[i].data_offset = h->data_len;
		h->data_len += BLOCKSIZE*DATA_BLOCK_SIZE;
		h->bucket[i].parity_offset = h->parity_len;
		h->bucket[i].last_parity_offset = h->parity_len;
		h->insertions++;
		*data_offset = h->bucket[i].data_offset;
		*parity_offset = h->bucket[i].parity_offset;
		
		if (rw)
		{
			h->bucket[i].write_count=1;
			*last_parity_offset = -1;
		}
		else
			h->bucket[i].read_count=1;
	} else {
		ext = h->bucket[i].next;
		ext_prev = &h->bucket[i];
		while(NULL!= ext)
		{
			if((int)key==ext->lba)
			{
				*data_offset = ext->data_offset;
				*parity_offset = ext->parity_offset;
				
				found = 1;
				_TRACE(TRACE_CACHE,"hash_insert : found int ext [%p]\n",ext);
				if (rw)
				{
					h->bucket[i].write_count++;
					*last_parity_offset = ext->last_parity_offset;
					ext->last_parity_offset = h->parity_len;
				}
				else
					h->bucket[i].read_count++;
				break;
			}
			ext_prev = ext;
			ext = ext->next;
		}
		if(!found)
		{
			ext = (PCDP_SUPER_BLOCK)malloc(sizeof(CDP_SUPER_BLOCK));
			ext->lba = key;
			ext->data_offset = h->data_len;
			ext->parity_offset = h->parity_len;
			ext->last_parity_offset = h->parity_len;
			ext->next = NULL;
			h->data_len += BLOCKSIZE*DATA_BLOCK_SIZE;
			h->collisions++;
			ext_prev->next = ext;
			TRACE_CDP("inserting key %u into %d->ext %p (collision)\n", key, i, ext);
			h->insertions++;
			*data_offset = ext->data_offset;
			*parity_offset = ext->parity_offset;
			
			if (rw)
			{
				h->bucket[i].write_count=1;
				*last_parity_offset = -1;
			}
			else
				h->bucket[i].read_count=1;
		}
	}
	return ISCSI_SUCCESS;
}
#else
int CDP_HashInsert(CDP_HASH_T *h, ULONG key, __int64 *data_offset, __int64 *parity_offset,BOOL rw)
{
	int i;
	PCDP_SUPER_BLOCK ext = NULL;
	PCDP_SUPER_BLOCK ext_prev = NULL;
	int found = 0;
	
	i = key/DATA_BLOCK_SIZE%(h->n);
	_TRACE(TRACE_CACHE,"hash_insert : key=%u, i=%d\n",key,i);
	if(h->bucket[i].lba==(int)key)
	{
		TRACE_CDP("hash_insert : found key=%u, i=%d\n",key,i);
		*data_offset = h->bucket[i].data_offset;
		*parity_offset = h->bucket[i].parity_offset;
		if (rw)
			h->bucket[i].write_count++;
		else
			h->bucket[i].read_count++;
		
		return ISCSI_SUCCESS;
	}
	if (h->bucket[i].lba==-1) {
		TRACE_CDP("inserting key %u  into %d,offset %d\n", key, i,h->data_len);
		h->bucket[i].lba = key;
		h->bucket[i].data_offset = h->data_len;
		h->data_len += BLOCKSIZE*DATA_BLOCK_SIZE;
		h->bucket[i].parity_offset = h->parity_len;
		h->bucket[i].archive_offset = -1;
		h->insertions++;
		*data_offset = h->bucket[i].data_offset;
		*parity_offset = h->bucket[i].parity_offset;
		if (rw)
			h->bucket[i].write_count=1;
		else
			h->bucket[i].read_count=1;
		
	} else {
		ext = h->bucket[i].next;
		ext_prev = &h->bucket[i];
		while(NULL!= ext)
		{
			if((int)key==ext->lba)
			{
				*data_offset = ext->data_offset;
				*parity_offset = ext->parity_offset;
				found = 1;
				_TRACE(TRACE_CACHE,"hash_insert : found int ext [%p]\n",ext);
				if (rw)
					h->bucket[i].write_count++;
				else
					h->bucket[i].read_count++;
				break;
			}
			ext_prev = ext;
			ext = ext->next;
		}
		if(!found)
		{
			ext = (PCDP_SUPER_BLOCK)malloc(sizeof(CDP_SUPER_BLOCK));
			ext->lba = key;
			ext->data_offset = h->data_len;
			ext->parity_offset = h->parity_len;
			ext->dirty = 0;
			ext->archive_offset = -1;			
			ext->next = NULL;
			h->data_len += BLOCKSIZE*DATA_BLOCK_SIZE;
			h->collisions++;
			ext_prev->next = ext;
			TRACE_CDP("inserting key %u into %d->ext %p (collision)\n", key, i, ext);
			h->insertions++;
			*data_offset = ext->data_offset;
			*parity_offset = ext->parity_offset;

			if (rw)
				ext->write_count=1;
			else
				ext->read_count=1;
			
		}
		
	}
	return ISCSI_SUCCESS;
}
#endif

#ifdef CDP_NO_GAP
int CDP_HashInquiry(CDP_HASH_T *h, ULONG key, __int64 *data_offset, __int64 *parity_offset, __int64 *last_parity_offset, BOOL rw)
{
	CDP_SUPER_BLOCK *curr;
	int i;
	
	i = key%(h->n);
	_TRACE(TRACE_CACHE,"hash_inquiry : key=%u, i=%d\n",key,i);
	curr = &h->bucket[i];
	if (curr->lba==(int)key) {
		TRACE_CDP("key %u  found in bucket[%d]\n", key, i);
		*data_offset = curr->data_offset;
		*parity_offset = curr->parity_offset;
		
		if (rw)
		{
			h->bucket[i].write_count++;
			*last_parity_offset = curr->last_parity_offset;
			curr->last_parity_offset = h->parity_len;
		}
		else
			h->bucket[i].read_count++;
		return ISCSI_SUCCESS;
	}
	while ((curr->lba!=(int)key)&&(curr->next!=NULL)) {
		curr = curr->next;
	}
	if (curr->lba!=(int)key) {
		TRACE_CDP("key %u not found in ext bucket\n", key);
		return ISCSI_ERROR;
	}
	else
	{
		TRACE_CDP("key %u found in ext bucket[%d]\n", key, i);
		*data_offset = curr->data_offset;
		*parity_offset = curr->parity_offset;
		
		if (rw)
		{
			h->bucket[i].write_count++;
			*last_parity_offset = curr->last_parity_offset;
			curr->last_parity_offset = h->parity_len;
		}
		else
			h->bucket[i].read_count++;
	}
	return ISCSI_SUCCESS;
	
}
#else
int CDP_HashInquiry(CDP_HASH_T *h, ULONG key, __int64 *data_offset, __int64 *parity_offset,BOOL rw,PCDP_SUPER_BLOCK *bucket)
{
	CDP_SUPER_BLOCK *curr;
	int i;
	
	i = key/DATA_BLOCK_SIZE%(h->n);
	_TRACE(TRACE_CACHE,"hash_inquiry : key=%u, i=%d\n",key,i);
	curr = &h->bucket[i];
	if (curr->lba==(int)key) {
		TRACE_CDP("key %u  found in bucket[%d]\n", key, i);
		*data_offset = curr->data_offset;
		*parity_offset = curr->parity_offset;
		if(bucket)
			*bucket = curr;
		if (rw)
			h->bucket[i].write_count++;
		else
			h->bucket[i].read_count++;
		return ISCSI_SUCCESS;
	}
	while ((curr->lba!=(int)key)&&(curr->next!=NULL)) {
		curr = curr->next;
	}
	if (curr->lba!=(int)key) {
		TRACE_CDP("key %u not found in ext bucket\n", key);
		if(bucket)
			*bucket = NULL;
		return ISCSI_ERROR;
	}
	else
	{
		TRACE_CDP("key %u found in ext bucket[%d]\n", key, i);
		*data_offset = curr->data_offset;
		*parity_offset = curr->parity_offset;
		if(bucket)
			*bucket = curr;
		if (rw)
			curr->write_count++;
		else
			curr->read_count++;
	}
	return ISCSI_SUCCESS;
	
}
#endif

int CDP_HashLoad(PCDP_HASH_T hash, int dev)
{
	int offset=0,next=0;
	int i;
	ULONG file_size_high=0,file_size_low=0;
	int n=0,collisions=0,insertions=0;
	__int64 data_len=0,parity_len=0,archive_len=0;
	unsigned int total_write_count;

	PCDP_SUPER_BLOCK curr,prev;

	if((NULL==hash)||(dev==0))
	{
		TRACE_ERROR("hash_load : Param invalid,hash=0x%p,dev=%d\n",hash,dev);
		return ISCSI_ERROR;
	}
	file_size_low=GetFileSize((HANDLE)dev,&file_size_high);
	if((file_size_low==0)&&(file_size_high==0))
	{
		_TRACE(TRACE_DEBUG,"hash_load : no need to load file\n");
		return ISCSI_SUCCESS;
	}
	offset += BE_Read_Byte(dev,(char*)&collisions,offset,4);
	offset += BE_Read_Byte(dev,(char*)&insertions,offset,4);
	offset += BE_Read_Byte(dev,(char*)&n,offset,4);
	offset += BE_Read_Byte(dev,(char*)&total_write_count,offset,4);
	offset += BE_Read_Byte(dev,(char*)&data_len,offset,8);
	offset += BE_Read_Byte(dev,(char*)&parity_len,offset,8);
#ifndef	CDP_NO_GAP
	offset += BE_Read_Byte(dev,(char*)&archive_len,offset,8);
#endif
	if(n==0)
	{
		return ISCSI_SUCCESS;
	}
	hash->collisions = collisions;
	hash->insertions = insertions;
	hash->n = n;
	hash->total_write_count = total_write_count;
	hash->data_len = data_len;
	hash->parity_len = parity_len;
#ifndef	CDP_NO_GAP
	hash->archive_len = archive_len;
#endif
	offset += BE_Read_Byte(dev,(char*)hash->bucket,offset,hash->n*sizeof(CDP_SUPER_BLOCK));
	for(i=0;i<hash->n;i++)
	{
		next = (int)hash->bucket[i].next;
		curr = &hash->bucket[i];
		while(next!=0)
		{
			prev = curr;
			if(NULL==(curr = (PCDP_SUPER_BLOCK)malloc(sizeof(CDP_SUPER_BLOCK))))
			{
				TRACE_ERROR("CDP_HashLoad : malloc curr failed");
				return ISCSI_ERROR;
			}
			BE_Read_Byte(dev,(char*)curr,next,sizeof(CDP_SUPER_BLOCK));
			next = (int)curr->next;
			curr->next = NULL;
			prev->next = curr;
		}
	}
	
	return ISCSI_SUCCESS;
}


int CDP_HashSave(PCDP_HASH_T hash, int dev)
{
	int offset=0,ext_offset=0;
	int i;
	PCDP_SUPER_BLOCK curr,prev;
	CDP_SUPER_BLOCK temp;
	
	if((NULL==hash)||(dev==0))
	{
		TRACE_ERROR("hash_save : Param invalid,hash=0x%p,dev=%d\n",hash,dev);
		return ISCSI_ERROR;
	}

	if (hash->insertions==0) {
		TRACE_ERROR("hash_save: don't need to save index file or no entry in hash table");
		return ISCSI_SUCCESS;
	}
	offset += BE_Write_Byte(dev,(char*)&hash->collisions,offset,4);
	offset += BE_Write_Byte(dev,(char*)&hash->insertions,offset,4);
	offset += BE_Write_Byte(dev,(char*)&hash->n,offset,4);
	offset += BE_Write_Byte(dev,(char*)&hash->total_write_count,offset,4);
	offset += BE_Write_Byte(dev,(char*)&hash->data_len,offset,8);
	offset += BE_Write_Byte(dev,(char*)&hash->parity_len,offset,8);
#ifndef	CDP_NO_GAP
	offset += BE_Write_Byte(dev,(char*)&hash->archive_len,offset,8);
#endif
	BE_Write_Byte(dev,(char *)hash->bucket,offset,hash->n*sizeof(CDP_SUPER_BLOCK));
	
	ext_offset = offset+hash->n*sizeof(CDP_SUPER_BLOCK);
	for(i=0;i<hash->n;i++)
	{
		curr = hash->bucket[i].next;
		prev = &hash->bucket[i];
		memcpy(&temp,prev,sizeof(CDP_SUPER_BLOCK ));
		if(curr)
		{
			temp.next = (PCDP_SUPER_BLOCK)ext_offset;
		}
		offset += BE_Write_Byte(dev,(char*)&temp,offset,sizeof(CDP_SUPER_BLOCK));
		while(curr)
		{
			prev = curr;
			curr = curr->next;
			memcpy(&temp,prev,sizeof(CDP_SUPER_BLOCK ));
			if(curr)
			{
				temp.next = (PCDP_SUPER_BLOCK)(ext_offset+sizeof(CDP_SUPER_BLOCK));
			}
			else
			{
				temp.next = 0;
			}
			ext_offset += BE_Write_Byte(dev,(char*)&temp,ext_offset,sizeof(CDP_SUPER_BLOCK));
		}
	}
	
	return ISCSI_SUCCESS;
	
}

