//#include "stdafx.h"
#include "backend.h"
#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <io.h>
#include <string.h>
#include "..\iscsi_lib\common\debug.h"

/*Function Name: BE_Init
  Description  : Open and initialize storage device
  Input        : Pointer of name buffer
  Output       : None
  Return       : Storage device ID
*/
int BE_Init(char* DevName)
{
	char szName[MAX_PATH];
	HANDLE hDevId;
	sprintf(szName, "%s", DevName);

	//iDevId=_open( szName, _O_RDWR | _O_CREAT | _O_BINARY, _S_IREAD |  _S_IWRITE );
	hDevId = CreateFile(szName, GENERIC_READ|GENERIC_WRITE, 
		FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if(hDevId==INVALID_HANDLE_VALUE)
	{
		hDevId = CreateFile(szName, GENERIC_READ|GENERIC_WRITE, 
			FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	}
	return (int)hDevId;
}

/*Function Name: BE_Close
  Description  : Close and release storage device
  Input        : Storage device ID
  Output       : None
  Return       : Status 0:success ,-1:error
*/
int BE_Close(int DevId)
{
	if(!DevId)
	{
		return ISCSI_SUCCESS;
	}
	//_close(DevId);
	CloseHandle((HANDLE)DevId);
	return ISCSI_SUCCESS;
}

//#ifdef OLD_BACKEND

/*Function Name: BE_Read
  Description  : Read data from storage device
  Input        : DevId - Storage device ID
                 Buff - Buffer to receive data
				 Lba - Start address
				 Length - Read length
  Output       : Buff - data
  Return       : >=0:length completed, -1:error
*/
int BE_Read(int DevId,char* Buff,unsigned long  Lba,unsigned long Length)
{
	LARGE_INTEGER Offset64;
	int iResult;
	//PRINT("BE_Read : lba=%d, length=%d\n",Lba,Length);
	Offset64.QuadPart = (__int64)Lba * BLOCKSIZE;
	
	//long Low  = (long)(Offset64&0x00000000ffffffff);
	//long High = (long)(Offset64>>32);
	DWORD dwCB=0;
	
	Offset64.LowPart=SetFilePointer((HANDLE)DevId, Offset64.LowPart, &Offset64.HighPart, FILE_BEGIN);
	if((Offset64.LowPart == 0xFFFFFFFF) && (GetLastError() != NO_ERROR))
	{
		TRACE_ERROR("BE_Read : SetFilePointer error");
		return ISCSI_ERROR;
	}

	iResult = ReadFile((HANDLE)DevId, Buff, Length, &dwCB, NULL);
	if(iResult!=0)
	{
		if(dwCB<Length)
		{
			//PRINT("BE_Read : lba=%d, length=%d\n",Lba,Length);
			//PRINT("BE_Read: dwCB=%d\n",dwCB);
			//MyGetLastError();
			memset(	Buff + dwCB,0,Length - dwCB);
		}
		iResult = Length;
	}
	else
	{
		TRACE_ERROR("ReadFile : Return 0\n");
		MyGetLastError();
		memset(Buff, 0,Length);
	}
	//MemDumpFunc(0,(UCHAR*)Buff,Length,"read result:\n");
	return iResult;
}
/*Function Name: BE_Write
  Description  : Write data to storage device
  Input        : DevId - Storage device ID
                 Buff - data Buffer 
				 Lba - Start address
				 Length - Write length
  Output       : Buff - data
  Return       : >=0:length completed, -1:error
*/
int BE_Write(int DevId,char* Buff,unsigned long  Lba,unsigned long Length)
{
	LARGE_INTEGER Offset64;
	int iResult;
	Offset64.QuadPart = (__int64)Lba * BLOCKSIZE;
	
	//long Low  = (long)(Offset64&0x00000000ffffffff);
	//long High = (long)(Offset64>>32);
	DWORD dwCB;
	
	Offset64.LowPart=SetFilePointer((HANDLE)DevId, Offset64.LowPart, &Offset64.HighPart, FILE_BEGIN);
	if((Offset64.LowPart == 0xFFFFFFFF) && (GetLastError() != NO_ERROR))
	{
		TRACE_ERROR("BE_Write : SetFilePointer error");
		return ISCSI_ERROR;
	}
	iResult = WriteFile((HANDLE)DevId, Buff, Length, &dwCB, NULL);
	//PRINT("BE_Write: iResult=%d\n",iResult);
	//MemDumpFunc(0,(UCHAR*)Buff,Length,"write result:\n");
	if(iResult)
	{
		if(dwCB<Length)
		{
			PRINT("BE_Write : lba=%d, length=%d\n",Lba,Length);
			PRINT("BE_Write: dwCB=%d\n",dwCB);
			MyGetLastError();
		}
		return Length;
	}
	else
	{
		PRINT("BE_Write : lba=%d, length=%d\n",Lba,Length);
		PRINT("BE_Write: dwCB=%d\n",dwCB);
		MyGetLastError();
		return iResult;
	}
	
}
int BE_Read_Byte(int DevId,char* Buff,__int64  Offset,unsigned long Length)
{
	__int64 Offset64;
	int iResult;
	//PRINT("BE_Read : lba=%d, length=%d\n",Lba,Length);
	Offset64 = (__int64)Offset;
	
	long Low  = (long)(Offset64&0x00000000ffffffff);
	long High = (long)(Offset64>>32);
	DWORD dwCB=0;
	
	SetFilePointer((HANDLE)DevId, Low, &High, FILE_BEGIN);
	iResult = ReadFile((HANDLE)DevId, Buff, Length, &dwCB, NULL);
	if(iResult!=0)
	{
		if(dwCB<Length)
		{
			memset(	Buff + dwCB,0,Length - dwCB);
		}
		iResult = Length;
	}
	if(iResult!=(int)Length)
	{
		TRACE_ERROR("BE_Read_Byte : Offset %d, expected %d, read %d",Offset,Length,iResult);
	}
	return iResult;
}
/*Function Name: BE_Write_Byte
  Description  : Write data to storage device
  Input        : DevId - Storage device ID
                 Buff - data Buffer 
				 Lba - Start address
				 Length - Write length
  Output       : Buff - data
  Return       : >=0:length completed, -1:error
*/
int BE_Write_Byte(int DevId,char* Buff,__int64  Offset,unsigned long Length)
{
	__int64 Offset64;
	int iResult;
	Offset64 = (__int64)Offset;
	
	long Low  = (long)(Offset64&0x00000000ffffffff);
	long High = (long)(Offset64>>32);
	DWORD dwCB;
	
	SetFilePointer((HANDLE)DevId, Low, &High, FILE_BEGIN);
	iResult = WriteFile((HANDLE)DevId, Buff, Length, &dwCB, NULL);
	//PRINT("BE_Write: iResult=%d\n",iResult);
	if(iResult)
	{
		return Length;
	}
	else
	{
		return iResult;
	}
	
}
int BE_Get_File_Length(int DevId)
{
	if(NULL==DevId)
	{
		TRACE_ERROR("BE_Get_File_Length : Param invalid\n");
		return 0;
	}
	return GetFileSize((HANDLE)DevId,NULL);
}