#include "snapshot.h"
#include "Storage.h"
#include "mythread.h"
#include "../BackEnd/backend.h"

//#define READ_IN_PIECES
//#define WRITE_IN_PIECES

int g_SnapshotIndexDev = 0;
int g_SnapshotDataDev = 0;
int g_OriginalDev = 0;
CDP_HASH_T *g_SnapshotHash = NULL;
CCriticalSection* g_SnapshotLock=NULL;
time_t gt_snapshot;

unsigned int sub_write_requests = 0;
unsigned int sub_read_requests = 0;
unsigned long total_write_size=0;
unsigned long total_read_size=0;

int Snapshot_Init(int dev)
{

	g_OriginalDev = dev;
	CTime t=CTime::GetCurrentTime();
	gt_snapshot = t.GetTime();
	if((int)INVALID_HANDLE_VALUE==(g_SnapshotIndexDev = BE_Init(SNAPSHOT_INDEX_FILE_NAME)))
	{
		TRACE_ERROR("Snapshot_Init : Open index file failed");
		return ISCSI_ERROR;
	}
	if((int)INVALID_HANDLE_VALUE==(g_SnapshotDataDev = BE_Init(SNAPSHOT_DATA_FILE_NAME)))
	{
		TRACE_ERROR("Snapshot_Init : Open data file failed");
		BE_Close(g_SnapshotIndexDev);
		return ISCSI_ERROR;
	}

	g_SnapshotHash = (PCDP_HASH_T)malloc(sizeof(CDP_HASH_T));
	if(ISCSI_SUCCESS!=CDP_HashInit(g_SnapshotHash,HASH_LENGTH))
	{
		TRACE_ERROR("Snapshot_Init : CDP_HashInit Failed");
		return ISCSI_ERROR;
	}
	if(ISCSI_SUCCESS!=CDP_HashLoad(g_SnapshotHash,g_SnapshotIndexDev))
	{
		TRACE_ERROR("CDP_Init : CDP_HashLoad Failed");
		return ISCSI_ERROR;
	}
	g_SnapshotLock = new CCriticalSection;
	return ISCSI_SUCCESS;
}

int Snapshot_Close()
{
	
	CDP_HashSave(g_SnapshotHash,g_SnapshotIndexDev);
	CDP_HashClose(g_SnapshotHash);
	free(g_SnapshotHash);
	BE_Close(g_SnapshotIndexDev);
	BE_Close(g_SnapshotDataDev);
	delete g_SnapshotLock;
	return ISCSI_SUCCESS;
}

int Redirect_Write(char *buf, unsigned long lba, unsigned long length)
{
	__int64 data_offset;
	__int64 parity_offset;
	__int64 last_parity_offset;
	ULONG blk_index;
	char *temp;
	ULONG lba_now,remain;
	ULONG to_write=0;
	ULONG count = 0;
	char *buf_now=buf;

	total_write_size += length/512;
	temp = (char*)malloc(DATA_BLOCK_SIZE*BLOCKSIZE);
	lba_now = lba;
	remain = lba&(DATA_BLOCK_SIZE-1);
	//PRINT("Write Request: lba=%d,length=%d\n",lba,length);
	g_SnapshotLock->Lock();
	if(remain)
	{
		lba_now = lba-remain;	
		if(ISCSI_SUCCESS==CDP_HashInquiry(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true))
		{
			if (length<=((DATA_BLOCK_SIZE-remain)*BLOCKSIZE)) {
				BE_Write(g_SnapshotDataDev,buf_now,(ULONG)(data_offset/BLOCKSIZE+remain),length);
				sub_write_requests++;
				goto out;
			}
			else {
				BE_Write(g_SnapshotDataDev,buf_now,(ULONG)(data_offset/BLOCKSIZE+remain),(DATA_BLOCK_SIZE-remain)*BLOCKSIZE);
				sub_write_requests++;
			}
		}
		else
		{
			BE_Read(g_OriginalDev,temp,lba_now,DATA_BLOCK_SIZE*BLOCKSIZE);
			CDP_HashInsert(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true);
			if (length<=((DATA_BLOCK_SIZE-remain)*BLOCKSIZE)) {
				memcpy(temp+remain*BLOCKSIZE,buf_now,length);
				sub_write_requests++;
				BE_Write(g_SnapshotDataDev,temp,(ULONG)(data_offset/BLOCKSIZE),DATA_BLOCK_SIZE*BLOCKSIZE);
				goto out;
			}
			else {
				memcpy(temp+remain*BLOCKSIZE,buf_now,(DATA_BLOCK_SIZE-remain)*BLOCKSIZE);
				BE_Write(g_SnapshotDataDev,temp,(ULONG)(data_offset/BLOCKSIZE),DATA_BLOCK_SIZE*BLOCKSIZE);
				sub_write_requests++;
			}

		}

		buf_now+=(DATA_BLOCK_SIZE-remain)*BLOCKSIZE;
		lba_now += DATA_BLOCK_SIZE;
		to_write = (DATA_BLOCK_SIZE-remain)*BLOCKSIZE;
	}

	count = (length - to_write)/(DATA_BLOCK_SIZE*BLOCKSIZE);
	remain = (length - to_write)%(DATA_BLOCK_SIZE*BLOCKSIZE);/*in BYTE*/
	
	for(blk_index=0;blk_index<count;blk_index++)
	{
		if(ISCSI_SUCCESS!=CDP_HashInquiry(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true))
		{
			CDP_HashInsert(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true);
		}

		BE_Write(g_SnapshotDataDev,buf_now,(ULONG)(data_offset/BLOCKSIZE),BLOCKSIZE*DATA_BLOCK_SIZE);
		sub_write_requests++;
		buf_now += BLOCKSIZE*DATA_BLOCK_SIZE;
		lba_now += DATA_BLOCK_SIZE;
	}

	if(remain<=0)
	{
		goto out;
	}

	if(ISCSI_SUCCESS==CDP_HashInquiry(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true))
	{
		BE_Write(g_SnapshotDataDev,buf_now,(ULONG)(data_offset/BLOCKSIZE),remain);
		sub_write_requests++;
		goto out;
	}
	else
	{
		BE_Read(g_OriginalDev,temp,lba_now,BLOCKSIZE*DATA_BLOCK_SIZE);
		CDP_HashInsert(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true);
		memcpy(temp,buf_now,remain);
		BE_Write(g_SnapshotDataDev,temp,(ULONG)(data_offset/BLOCKSIZE),BLOCKSIZE*DATA_BLOCK_SIZE);
		sub_write_requests++;
	}
out:
	free(temp);
	g_SnapshotLock->Unlock();
	return ISCSI_SUCCESS;
}

int Merge_Read(char *buf, unsigned long lba, unsigned long length)
{
	__int64 data_offset;
	__int64 parity_offset;
	ULONG blk_index;
	char *temp;
	ULONG lba_now,remain;
	ULONG to_write=0;
	int to_read=-1;
	ULONG count = 0;
	char *buf_now=buf;

	total_read_size += length/512; 
	if (g_SnapshotHash==NULL) {
		BE_Read(g_OriginalDev,buf,lba,length);
		return ISCSI_SUCCESS;
	}
	
	temp = (char*)malloc(DATA_BLOCK_SIZE*BLOCKSIZE);

	lba_now = lba;
	remain = lba&(DATA_BLOCK_SIZE-1);
	
	//PRINT("Read Request: lba=%d,length=%d\n",lba,length);
	g_SnapshotLock->Lock();
	if(remain)
	{
		lba_now = lba-remain;					
		if(ISCSI_SUCCESS==CDP_HashInquiry(g_SnapshotHash,lba_now,&data_offset,&parity_offset))
		{
			BE_Read(g_SnapshotDataDev,temp,(ULONG)(data_offset/BLOCKSIZE),BLOCKSIZE*DATA_BLOCK_SIZE);
			sub_read_requests++;
		}
		else
		{
			BE_Read(g_OriginalDev,temp,lba_now,BLOCKSIZE*DATA_BLOCK_SIZE);
			sub_read_requests++;
		}

		to_read = (DATA_BLOCK_SIZE-remain)*BLOCKSIZE-length;
		to_write =  (DATA_BLOCK_SIZE-remain)*BLOCKSIZE;

		if(to_read>0)
		{
			to_write = length;
		}

		memcpy(buf_now,temp+remain*BLOCKSIZE,to_write);

		buf_now += to_write;
		lba_now += DATA_BLOCK_SIZE;
	}

	if(to_read>=0)
	{
		goto out;
	}
	
	count = (length - to_write)/(DATA_BLOCK_SIZE*BLOCKSIZE);
	remain = (length - to_write)%(DATA_BLOCK_SIZE*BLOCKSIZE);/*in BYTE*/
	
	for(blk_index=0;blk_index<count;blk_index++)
	{
		if(ISCSI_SUCCESS==CDP_HashInquiry(g_SnapshotHash,lba_now,&data_offset,&parity_offset))
		{
			BE_Read(g_SnapshotDataDev,temp,(ULONG)(data_offset/BLOCKSIZE),BLOCKSIZE*DATA_BLOCK_SIZE);
			sub_read_requests++;
		}
		else
		{
			BE_Read(g_OriginalDev,temp,lba_now,BLOCKSIZE*DATA_BLOCK_SIZE);
			sub_read_requests++;
		}
		memcpy(buf_now,temp,BLOCKSIZE*DATA_BLOCK_SIZE);
		buf_now += BLOCKSIZE*DATA_BLOCK_SIZE;
		lba_now += DATA_BLOCK_SIZE;
	}

	if(remain<=0)
	{
		goto out;
	}

	if(ISCSI_SUCCESS==CDP_HashInquiry(g_SnapshotHash,lba_now,&data_offset,&parity_offset))
	{
		BE_Read(g_SnapshotDataDev,temp,(ULONG)(data_offset/BLOCKSIZE),BLOCKSIZE*DATA_BLOCK_SIZE);
		sub_read_requests++;
	}
	else
	{
		BE_Read(g_OriginalDev,temp,lba_now,BLOCKSIZE*DATA_BLOCK_SIZE);
		sub_read_requests++;
	}
	memcpy(buf_now,temp,remain);

out:
	free(temp);
	g_SnapshotLock->Unlock();
	return ISCSI_SUCCESS;
}

int Merge_Snapshot(long t_snapshot)
{
	int i;
	PCDP_SUPER_BLOCK bucket;
	char *temp;
	__int64 data_offset;

	if ((g_SnapshotHash==NULL)||(g_SnapshotHash->insertions==0)) {
		PRINT("Don't need to merge");
		return ISCSI_SUCCESS;
	}

	temp = (char*)malloc(DATA_BLOCK_SIZE*BLOCKSIZE);

	g_SnapshotLock->Lock();
	for(i=0;i<g_SnapshotHash->n;i++)
	{
		bucket = &g_SnapshotHash->bucket[i];
		if(bucket->lba==-1)
			continue;
		while(bucket)
		{

			data_offset = bucket->data_offset;
			BE_Read(g_SnapshotDataDev,temp,(ULONG)(data_offset/BLOCKSIZE),BLOCKSIZE*DATA_BLOCK_SIZE);
			BE_Write(g_OriginalDev,temp,bucket->lba,BLOCKSIZE*DATA_BLOCK_SIZE);
			bucket = bucket->next;
		}
	}
	free(temp);

	CDP_HashSave(g_SnapshotHash,g_SnapshotIndexDev);
	CDP_HashClose(g_SnapshotHash);
	BE_Close(g_SnapshotDataDev);
	BE_Close(g_SnapshotIndexDev);
	free(g_SnapshotHash);
	
	CTime t = CTime(t_snapshot);
	char snapshotName[40];
	sprintf(snapshotName,"SNAPSHOT1_%s.dat",t.Format("%Y%m%d%H%M%S"));
	MoveFile(SNAPSHOT_DATA_FILE_NAME,snapshotName);
	sprintf(snapshotName,"SNAPSHOT1_%s.idx",t.Format("%Y%m%d%H%M%S"));
	MoveFile(SNAPSHOT_INDEX_FILE_NAME,snapshotName);

	if((int)INVALID_HANDLE_VALUE==(g_SnapshotIndexDev = BE_Init(SNAPSHOT_INDEX_FILE_NAME)))
	{
		TRACE_ERROR("Snapshot_Init : Open index file failed");
		return ISCSI_ERROR;
	}
	if((int)INVALID_HANDLE_VALUE==(g_SnapshotDataDev = BE_Init(SNAPSHOT_DATA_FILE_NAME)))
	{
		TRACE_ERROR("Snapshot_Init : Open data file failed");
		BE_Close(g_SnapshotIndexDev);
		return ISCSI_ERROR;
	}

	g_SnapshotHash = (PCDP_HASH_T)malloc(sizeof(CDP_HASH_T));
	if(ISCSI_SUCCESS!=CDP_HashInit(g_SnapshotHash,HASH_LENGTH))
	{
		TRACE_ERROR("Snapshot_Init : CDP_HashInit Failed");
		return ISCSI_ERROR;
	}
	if(ISCSI_SUCCESS!=CDP_HashLoad(g_SnapshotHash,g_SnapshotIndexDev))
	{
		TRACE_ERROR("CDP_Init : CDP_HashLoad Failed");
		return ISCSI_ERROR;
	}

	g_SnapshotLock->Unlock();
	return ISCSI_SUCCESS;

}

int Copy_on_Write(char *buf, unsigned long lba, unsigned long length)
{
	__int64 data_offset;
	__int64 parity_offset;
	__int64 last_parity_offset;
	ULONG blk_index;
	char *temp;
	ULONG lba_now,remain;
	ULONG to_write=0;
	ULONG count = 0;
	char *buf_now=buf;

	total_write_size += length/512;
	temp = (char*)malloc(DATA_BLOCK_SIZE*BLOCKSIZE);
	lba_now = lba;
	remain = lba&(DATA_BLOCK_SIZE-1);
	//PRINT("Write Request: lba=%d,length=%d\n",lba,length);
	g_SnapshotLock->Lock();
	if(remain)
	{
		lba_now = lba-remain;
		if(ISCSI_SUCCESS!=CDP_HashInquiry(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true))
		{
			CDP_HashInsert(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true);
			BE_Read(g_OriginalDev,temp,lba_now,DATA_BLOCK_SIZE*BLOCKSIZE);
			BE_Write(g_SnapshotDataDev,temp,(ULONG)(data_offset/BLOCKSIZE),DATA_BLOCK_SIZE*BLOCKSIZE);
		}
		
		if (length<=((DATA_BLOCK_SIZE-remain)*BLOCKSIZE))
		{
#ifdef WRITE_IN_PIECES
			BE_Write(g_OriginalDev,buf_now,lba_now+remain,length);
#endif
			sub_write_requests++;
			goto out;
		}
		
		sub_write_requests++;
#ifdef WRITE_IN_PIECES
		BE_Write(g_OriginalDev,buf_now,lba_now+remain,(DATA_BLOCK_SIZE-remain)*BLOCKSIZE);
#endif
		buf_now+=(DATA_BLOCK_SIZE-remain)*BLOCKSIZE;
		lba_now += DATA_BLOCK_SIZE;
		to_write = (DATA_BLOCK_SIZE-remain)*BLOCKSIZE;
	}

	count = (length - to_write)/(DATA_BLOCK_SIZE*BLOCKSIZE);
	remain = (length - to_write)%(DATA_BLOCK_SIZE*BLOCKSIZE);/*in BYTE*/
	
	for(blk_index=0;blk_index<count;blk_index++)
	{
		if(ISCSI_SUCCESS!=CDP_HashInquiry(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true))
		{
			CDP_HashInsert(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true);
			BE_Read(g_OriginalDev,temp,lba_now,DATA_BLOCK_SIZE*BLOCKSIZE);
			BE_Write(g_SnapshotDataDev,temp,(ULONG)(data_offset/BLOCKSIZE),DATA_BLOCK_SIZE*BLOCKSIZE);
		}
		sub_write_requests++;
#ifdef WRITE_IN_PIECES
		BE_Write(g_OriginalDev,buf_now,lba_now,DATA_BLOCK_SIZE*BLOCKSIZE);
#endif
		buf_now += BLOCKSIZE*DATA_BLOCK_SIZE;
		lba_now += DATA_BLOCK_SIZE;
	}

	if(remain<=0)
	{
		goto out;
	}

	if(ISCSI_SUCCESS!=CDP_HashInquiry(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true))
	{
		CDP_HashInsert(g_SnapshotHash,lba_now,&data_offset,&parity_offset,&last_parity_offset,true);
		BE_Read(g_OriginalDev,temp,lba_now,DATA_BLOCK_SIZE*BLOCKSIZE);
		BE_Write(g_SnapshotDataDev,temp,(ULONG)(data_offset/BLOCKSIZE),DATA_BLOCK_SIZE*BLOCKSIZE);
	}
	sub_write_requests++;
#ifdef WRITE_IN_PIECES
	BE_Write(g_OriginalDev,buf_now,lba_now,remain);
#endif
out:

#ifndef WRITE_IN_PIECES
	BE_Write(g_OriginalDev,buf,lba,length);
#endif
	free(temp);
	g_SnapshotLock->Unlock();
	return ISCSI_SUCCESS;
}

int Snapshot_Read(char *buf, unsigned long lba, unsigned long length)
{
	ULONG blk_index;
	ULONG lba_now;
	ULONG remain,to_write=0;
	ULONG count = 0;
	char *buf_now=buf;

	total_read_size += length/512;

#ifdef READ_IN_PIECES
	lba_now = lba;
	remain = lba&(DATA_BLOCK_SIZE-1);
	//PRINT("Write Request: lba=%d,length=%d\n",lba,length);
	g_SnapshotLock->Lock();
	if(remain)
	{
		lba_now = lba-remain;			
		
		if (length<=((DATA_BLOCK_SIZE-remain)*BLOCKSIZE))
		{
			sub_read_requests++;
			BE_Read(g_OriginalDev,buf_now,lba_now+remain,length);
			goto out;
		}
		sub_read_requests++;
		BE_Read(g_OriginalDev,buf_now,lba_now+remain,(DATA_BLOCK_SIZE-remain)*BLOCKSIZE);
		buf_now+=(DATA_BLOCK_SIZE-remain)*BLOCKSIZE;
		lba_now += DATA_BLOCK_SIZE;
		to_write = (DATA_BLOCK_SIZE-remain)*BLOCKSIZE;
	}

	count = (length - to_write)/(DATA_BLOCK_SIZE*BLOCKSIZE);
	remain = (length - to_write)%(DATA_BLOCK_SIZE*BLOCKSIZE);/*in BYTE*/
	
	for(blk_index=0;blk_index<count;blk_index++)
	{
		sub_read_requests++;
		BE_Read(g_OriginalDev,buf_now,lba_now,DATA_BLOCK_SIZE*BLOCKSIZE);
		buf_now += BLOCKSIZE*DATA_BLOCK_SIZE;
		lba_now += DATA_BLOCK_SIZE;
	}

	if(remain<=0)
	{
		goto out;
	}

	sub_read_requests++;
	BE_Read(g_OriginalDev,buf_now,lba_now,remain);
out:

	g_SnapshotLock->Unlock();
#else
	long to_read=0;
	lba_now = lba;
	remain = lba&(DATA_BLOCK_SIZE-1);
	if (remain) {
		sub_read_requests++;
		to_read = length-(DATA_BLOCK_SIZE-remain)*BLOCKSIZE;
	}
	else
		to_read = length;
	if (to_read>0) {
		sub_read_requests+=to_read/(DATA_BLOCK_SIZE*BLOCKSIZE)+(((to_read%(DATA_BLOCK_SIZE*BLOCKSIZE))>0)?1:0);
	}

	BE_Read(g_OriginalDev,buf,lba,length);
#endif
	return ISCSI_SUCCESS;
}

int measure_rw()
{
	int i,j;
	char buffer[64*1024];
	ULONG lba;
	memset(buffer,0,64*1024);
	srand( (unsigned)time( NULL ) );
	lba = rand();
	clock_t t1,t2;
/*
	t1=clock();
	for (i=0;i<1000;i++)
		BE_Write(g_SnapshotDataDev,buffer,lba,64*1024);
	t2=clock();
	PRINT("BSIZE=64k Write 64k buffer 1000 time =%.3f\n",(float)(t2-t1)/CLOCKS_PER_SEC);

	lba=rand();
	t1=clock();
	for (i=0;i<1000;i++)
		BE_Read(g_SnapshotDataDev,buffer,lba,64*1024);
	t2=clock();
	PRINT("BSIZE=64k Read 64k buffer 1000 time =%.3f\n",(float)(t2-t1)/CLOCKS_PER_SEC);
*/

	lba=rand();
	t1 = clock();
	for (i=0;i<1000;i++)
		for (j=0;j<128;j++)
			BE_Write(g_SnapshotDataDev,buffer+j*512,lba+j,512);
	t2 = clock();
	PRINT("BSIZE=.5k Write 64k buffer 1000 time =%.3f\n",(float)(t2-t1)/CLOCKS_PER_SEC);

	lba=rand();
	t1 = clock();
	for (i=0;i<1000;i++)
		for (j=0;j<128;j++)
			BE_Read(g_SnapshotDataDev,buffer+j*512,lba+j,512);
	t2 = clock();
	PRINT("BSIZE=.5k Read 64k buffer 1000 time =%.3f\n",(float)(t2-t1)/CLOCKS_PER_SEC);
/*
	lba=rand();
	t1 = clock();
	for (i=0;i<1000;i++)
		for (j=0;j<16;j++)
			BE_Write(g_SnapshotDataDev,buffer+j*4096,lba+j*8,4096);
	t2 = clock();
	PRINT("BSIZE=4k Write 64k buffer 1000 time =%.3f\n",(float)(t2-t1)/CLOCKS_PER_SEC);

	lba=rand();
	t1 = clock();
	for (i=0;i<1000;i++)
		for (j=0;j<16;j++)
			BE_Read(g_SnapshotDataDev,buffer+j*4096,lba+j*8,4096);
	t2 = clock();
	PRINT("BSIZE=4k Read 64k buffer 1000 time =%.3f\n",(float)(t2-t1)/CLOCKS_PER_SEC);

	lba=rand();
	t1 = clock();
	for (i=0;i<1000;i++)
		for (j=0;j<8;j++)
			BE_Write(g_SnapshotDataDev,buffer+j*8192,lba+j*16,8192);
	t2 = clock();
	PRINT("BSIZE=8k Write 64k buffer 1000 time =%.3f\n",(float)(t2-t1)/CLOCKS_PER_SEC);

	lba=rand();
	t1 = clock();
	for (i=0;i<1000;i++)
		for (j=0;j<8;j++)
			BE_Read(g_SnapshotDataDev,buffer+j*8192,lba+j*16,8192);
	t2 = clock();
	PRINT("BSIZE=8k Read 64k buffer 1000 time =%.3f\n",(float)(t2-t1)/CLOCKS_PER_SEC);


	lba=rand();
	t1 = clock();
	for (i=0;i<1000;i++)
		for (j=0;j<4;j++)
			BE_Write(g_SnapshotDataDev,buffer+j*16*1024,lba+j*32,16*1024);
	t2 = clock();
	PRINT("BSIZE=16k Write 64k buffer 1000 time =%.3f\n",(float)(t2-t1)/CLOCKS_PER_SEC);

	lba=rand();
	t1 = clock();
	for (i=0;i<1000;i++)
		for (j=0;j<4;j++)
			BE_Read(g_SnapshotDataDev,buffer+j*16*1024,lba+j*32,16*1024);
	t2 = clock();
	PRINT("BSIZE=16k Read 64k buffer 1000 time =%.3f\n",(float)(t2-t1)/CLOCKS_PER_SEC);

*/
	return ISCSI_SUCCESS;

}

int New_Snapshot(long t_snapshot)
{

	measure_rw();
	return ISCSI_SUCCESS;
/*	g_SnapshotLock->Lock();
	char snapshotName[40];
	CDP_HashSave(g_SnapshotHash,g_SnapshotIndexDev);
	CDP_HashClose(g_SnapshotHash);
	free(g_SnapshotHash);
	BE_Close(g_SnapshotDataDev);
	BE_Close(g_SnapshotIndexDev);
	CTime t = CTime(gt_snapshot);
	gt_snapshot = t_snapshot;
	sprintf(snapshotName,"SNAPSHOT2_%s.dat",t.Format("%Y%m%d%H%M%S"));
	MoveFile(SNAPSHOT_DATA_FILE_NAME,snapshotName);
	sprintf(snapshotName,"SNAPSHOT2_%s.idx",t.Format("%Y%m%d%H%M%S"));
	MoveFile(SNAPSHOT_INDEX_FILE_NAME,snapshotName);
	
	if((int)INVALID_HANDLE_VALUE==(g_SnapshotIndexDev = BE_Init(SNAPSHOT_INDEX_FILE_NAME)))
	{
		TRACE_ERROR("Snapshot_Init : Open index file failed");
		return ISCSI_ERROR;
	}
	if((int)INVALID_HANDLE_VALUE==(g_SnapshotDataDev = BE_Init(SNAPSHOT_DATA_FILE_NAME)))
	{
		TRACE_ERROR("Snapshot_Init : Open data file failed");
		BE_Close(g_SnapshotIndexDev);
		return ISCSI_ERROR;
	}

	g_SnapshotHash = (PCDP_HASH_T)malloc(sizeof(CDP_HASH_T));
	if(ISCSI_SUCCESS!=CDP_HashInit(g_SnapshotHash,HASH_LENGTH))
	{
		TRACE_ERROR("Snapshot_Init : CDP_HashInit Failed");
		return ISCSI_ERROR;
	}
	if(ISCSI_SUCCESS!=CDP_HashLoad(g_SnapshotHash,g_SnapshotIndexDev))
	{
		TRACE_ERROR("CDP_Init : CDP_HashLoad Failed");
		return ISCSI_ERROR;
	}	
	
	g_SnapshotLock->Unlock();
	return ISCSI_SUCCESS;
*/
}